/* ThermometerOne, Plan "B" ATMEGA328 Version */
/* Release 1.0.0, October 2013, Public Domain */

#include <avr/sleep.h>                      // needed for shutdown function
#include <EEPROM.h>                         // needed for EEPROM read and write
#include <HexDecAsc.h>                      // used for EEPROM dump All

// EEPROM Address Constants
const word EEmask   =  0;                   //  1 byte location of EEPROM storage mode mask
const word EEflag   =  1;                   //  1 byte location of EEPROM storage mode flag
const word EErefvolt=  2;                   //  2 byte location of RefVoltage 
const word EEoffset =  4;                   //  2 byte location of DegreeOffset 
const word EEminutes=  6;                   //  2 byte location of Report Target Minutes
const word EEunused0=  8;                   //  2 byte location -- reserved -- unused
const word EEunused1= 10;                   //  2 byte location -- reserved -- unused
const word EEunused2= 12;                   //  2 byte location -- reserved -- unused
const word EEunused3= 14;                   //  2 byte location -- reserved -- unused
const word EEidtring= 16;                   // ID string w/o termiantion size (16) 
const word EEidsize = 16;                   // 24 byte location of IdString 
const word EEwdsize = EEidtring+EEidsize;   // Working data storage size      (32)
//-----------------------------  ---- ----  // -- This area reserved for Table based system --
const word EEtable  = 32;		    // 40 word location of converion table
const word EEtbsize = 80;                   // Working data storage size      (80)
//-----------------------------  ---- ----  // -- This area reserved for Table based system --
const word StorageWorking=EEmask;           // EEPROM start for working calibration data
// EEPROM start for backup  copy of constants 
// note we have to add 1 to the value
// Becasue addresses begin with zero not one
const word StorageBackup =((E2END-(EEwdsize))+1);       

// EEPROM addresses variables               // begin storage for report data
word  StorageBegin =StorageWorking+EEwdsize+EEtbsize;
word  StorageMark  =StorageBegin;           // marks start of current segment
word  StorageEnd   =StorageBackup;          // marks end of current segment
word  StorageIndex =StorageBegin;           // index for next EEPROM write

// Conversion Factors/Calibraton Data
float RefVoltage;                           // Analog Sensor Reference Voltage
const float CovrtFactorV= 8192;
float DegreeOffset;                         // Fudge factor to adjust output
const float CovrtFactorO= 1024;             
char  IdString[EEidsize+1];                 // ID/Location string for this device
word  MinuteTarget      = 1;                // Number of minutes between report lines
byte  newflg=0;                             // used to indicate new factors in memory

float Celsius;                              // Last conversion to Celsius Temperature
float Fahrenheit;                           // Last conversion to Fahrenheit Temperature
float AvrCelsius;                           // Last AVR conversion to Celsius Temperature
float AVRFahrenheit;                        // Last AVR conversion to Fahrenheit Temperature

// Global operational mode Variables        // set default operation modes
boolean       ReportMode    = true;         // True =  reporting, False = Command Mode 
boolean       RtnRawRead    = true;         // True =  include Raw
boolean       RtnCelsius    = true;         // True =  include Celsius
boolean       RtnFahrenh    = true;         // True =  include Fahrenheit
boolean       DeBug         = false;        // True =  extended reporting for debugging
boolean       RtnAvrRead    = false;        // True =  include Avr internal Temperature line
boolean       EepromMode    = false;        // True =  write data to EEPROM on next run
boolean       RoundMode     = true;         // True =  round output to nearest 1/4 or 1/2

// Global work Variables                    // There are 1000 milliseconds in a second
// word        SecondsMinute = 10000;       // --- to speed things up a bit for debugging
word          SecondsMinute = 60000;        // added so calibraton timining can be reduced
unsigned long SecondsTarget = 0;            // Number of seconds between report lines
unsigned long RptStartTime  = 0;            // Time between report lines 
byte          gap           = 0;            // used to increase gap between reads
unsigned long RptTrigger    = 0;            // Target Time for report 
unsigned long CycleStart    = 0;            // Target Time for report 
unsigned long CycleTime     = 0;            // Target Time for report 

unsigned long Accumalator   = 0;            // Accumalate temperature reads
unsigned long CycleCount    = 0;            // Cycles per Report line
char          cmd[]         = {0,0,0};      // used to store two character command
char          prevcmd[]     = {0,0,0};      // used to store previous two character command

word          LastRead      = 0;            // Stores previous RawRead Average
byte          Consecutive   = 0;            // used to count consective equal readings
                                            
//=====================================================================================
void setup() 
  { char c;
    Serial.begin (9600);
    pinMode(13, OUTPUT);                     // so we can blink it later during writes
    EnableADC();                             // enables the ADC and set ADC clock factor
    delay (1000);                            // let serial library complete setup
    while (Serial.available()>0)             // drain any data from the serial buffer 
        c=Serial.read();
    Read_Calibration_Data();                 // read and set conversion factors from EEPROM
    Check_EEPROM();                          // see if we are writing to EEPROM vs Serial
    // calculate seconds between report lines
    // we have to "cast" the two word values or we will get a word value for the result
    SecondsTarget=long(MinuteTarget)*long(SecondsMinute);
    //delay (5000);                          // allow PC 5 seconds to get setup
    if (EepromMode==false) ReportStatus();   // report default parameters
    Accumalator = 0;                         // set startup parameters
    CycleCount  = 0;
    RptTrigger  = millis() + SecondsTarget;
    RptStartTime= millis(); 
    //------debugging stuff--------------------
    // Serial.println ("Got here");
    // while (true);
  }

//-----------------------------------------------------------
void loop() 
  { char c1, c2;
    word wtemp;
    
    // ---- This is where we check for command input
    if ((Serial.available()>2) && (EepromMode==false))
      { if(ReadTwoCharacters()) CmdProcessor();}
     
    // ---- This is where we collect our temperature data
    // gap is used to increase the amount of time between reading sampling the ADC.
    // This Insures that we will not miss any data transmitted on the serial port.
    if (gap++ == 9)           
      { gap=0;                     
        // cycle times are only used if debugging is turned on
        if (DeBug == true) CycleStart= millis(); 
        ReadRawTempA1();
        if (DeBug == true) CycleTime = CycleTime+(millis()-CycleStart);
      }
    
    // ---- This is where we output the teperature data
    // The time required to read 64 samples is about 119-120 milliseconds. If we get
    // within 150 milliseconds of the Report Trigger Time then we wait for it.
    // With the these timing numbers there are 500 reads of 
    // 64 virtual 12 bit samples per minute.
    // Added condition for millis exceeding report trigger (possible with long commands)    
    if (((RptTrigger-millis())< 150) || (millis()>RptTrigger))
      { // Serial.println ("Got here: RptTrigger ");
        while (millis() < RptTrigger);
        // We want the new trigger time set as close as possible to when the previous trigger 
        // went off --- so we put ti first.
        RptTrigger= (millis() + (SecondsTarget));
        // Serial.print (F("Got Here: RptTrigger, milliseconds to wait= "));    
        // Serial.println (SecondsTarget);
        Report();
      }
  } 

//================================================================  
void CmdProcessor()
  {// this function is the main command handler
   // not many comments because I think the code is obviuos
   if (DeBug == true)
       { Serial.print (F("; Command Processor "));
         DebugPrintCharacters (cmd[0],cmd[1]);
        }
    if      ((cmd[0]=='C') && (cmd[1]=='=')) CelsiusEquals();
    else if ((cmd[0]=='D') && (cmd[1]=='B')) ToggleDebugMode();
    else if ((cmd[0]=='D') && (cmd[1]=='O')) NewDegreeOffset();
    else if ((cmd[0]=='D') && (cmd[1]=='0')) NewDegreeOffset();  // for typoes
    else if ((cmd[0]=='E') && (cmd[1]=='+')) EEmodeFlagSet();
    else if ((cmd[0]=='E') && (cmd[1]=='C')) ClearStorage();
    else if ((cmd[0]=='E') && (cmd[1]=='D')) DumpStorage();
    else if ((cmd[0]=='E') && (cmd[1]=='-')) EEmodeFlagClear();
    else if ((cmd[0]=='F') && (cmd[1]=='=')) FahrenheitEquals();
    else if ((cmd[0]=='I') && (cmd[1]=='D')) {Serial.print(F("; ")); Print_IdString();}
    else if ((cmd[0]=='L') && (cmd[1]=='L')) HelpMe();
    else if ((cmd[0]=='L') && (cmd[1]==':')) NewIdString();
    else if ((cmd[0]=='R') && (cmd[1]=='V')) NewRefVolt();
    else if ((cmd[0]=='S') && (cmd[1]=='S')) ShutDown();
    else if ((cmd[0]=='S') && (cmd[1]=='T')) ReportStatus();
    else if ((cmd[0]=='W') && (cmd[1]=='W')) Write_Calibration_Data();
    else if ((cmd[0]=='W') && (cmd[1]=='+')) OverwriteBackup();
    else if ((cmd[0]=='W') && (cmd[1]=='-')) RestoreFromBackup();
    else if ((cmd[0]=='?') && (cmd[1]=='?')) HelpMe();
    else if ((cmd[0]=='!') && (cmd[1]=='!')) software_Reset();
    else if ((cmd[0]=='0') && (cmd[1]=='0')) ToggleRoundMode();
    
    else if  (cmd[0]=='C')                   SetCelsiusMode();
    else if  (cmd[0]=='F')                   SetFahrenheitdMode();
    else if  (cmd[0]=='I')                   SetAvrInternalMode();
    else if  (cmd[0]=='P')                   SetReportMode();
    else if  (cmd[0]=='R')                   SetRawReadMode();
    else if  (cmd[0]=='T')                   NewReportTime();
    
    // example of commands not implemented
    else if ((cmd[0]=='A') && (cmd[1]==':')) PrintNotImplemented();
    else if ((cmd[0]=='S') && (cmd[1]==':')) PrintNotImplemented();
    
    // example of application specific command implimneted
    // these two commands write test data to the EEPROM working storage
    else if ((cmd[0]=='Z') && (cmd[1]=='1')) TestData1();
    else if ((cmd[0]=='Z') && (cmd[1]=='2')) TestData2();
    // this command used for calibration, changes reporting to 5 seconds
    else if ((cmd[0]=='Z') && (cmd[1]=='Z')) CalibrationMode();
    // this command used to dump entire EEPROM to Serail Port
    else if ((cmd[0]=='Z') && (cmd[1]=='D')) EepromDumpAll();
    else PrintNotRecognized();               // not recognized
    // ----------------------------------------------------------
    // previously define commands not valid in this implimentation
    // else if ((cmd[0]=='O') && (cmd[1]==':')) NewOffsetR();
    // else if ((cmd[0]=='F') && (cmd[1]==':')) NewFahrenheit();
    // else if ((cmd[0]=='C') && (cmd[1]==':')) NewCelsius();
    // else if ((cmd[0]=='C') && (cmd[1]=='=')) CelsiusEquals();
    // else if ((cmd[0]=='T') && (cmd[1]=='T')) TestTest();

  }

//---------------------------------------------------------------------- 
//void TestTest()
//  { for (byte i=0; i< 20; i++)
//    { QuickBlink();
//      delay (200);
//    }
//  }

//---------------------------------------------------------------------- 
void HelpMe()
  //Serial.println(F("This string will be stored in flash memory"));
  { PrintSeperatorLine();
    Serial.println(F(";\t\tArduino AtMega328 Temperature Sensor 1.0.0"));
    Serial.println(F(";\tID\tOutput ID string"));
    Serial.println(F(";\tST\tOutput Status"));
    Serial.println(F(";\tRT\tRaw=True"));
    Serial.println(F(";\tRF\tRaw=False"));
    Serial.println(F(";\tFT\tFahrenheit=True"));
    Serial.println(F(";\tFF\tFahrenheit=False"));
    Serial.println(F(";\tF=\tEnter Current Fahrenheit"));
    Serial.println(F(";\tCT\tCelsius=True"));
    Serial.println(F(";\tCF\tCelsius=False"));
    Serial.println(F(";\tIT\tAVR Internal Temperature=True"));
    Serial.println(F(";\tIF\tAVR Internal Temperature=False"));
    Serial.println(F(";\tC=\tEnter Current Celsius"));  // to be done
    Serial.println(F(";\tDO\tNew Degree Offset (Fahrenheit)"));
    Serial.println(F(";\tDF\tSame as DO"));
    Serial.println(F(";\tRV\tNew Reference Voltage"));
    Serial.println(F(";\tT1\tReport time = 01 minutes"));
    Serial.println(F(";\tT2\tReport time = 02 minutes"));
    Serial.println(F(";\tT3\tReport time = 03 minutes"));
    Serial.println(F(";\tT4\tReport time = 04 minutes"));
    Serial.println(F(";\tT5\tReport time = 05 minutes"));
    Serial.println(F(";\tT6\tReport time = 10 minutes"));
    Serial.println(F(";\tT7\tReport time = 15 minutes"));
    Serial.println(F(";\tT8\tReport time = 20 minutes"));
    Serial.println(F(";\tT9\tReport time = 30 minutes"));
    Serial.println(F(";\tT0\tReport time = 60 minutes"));
    Serial.println(F(";\tTA\tReport time = 02 hours"));
    Serial.println(F(";\tTB\tReport time = 04 hours"));
    Serial.println(F(";\tTC\tReport time = 06 hours"));
    Serial.println(F(";\tTD\tReport time = 08 hours"));
    Serial.println(F(";\tTE\tReport time = 12 hours"));
    Serial.println(F(";\tTF\tReport time = 24 hours"));
    Serial.println(F(";\tPF\tPrint mode = False"));
    Serial.println(F(";\tPT\tPrint mode = True"));
    Serial.println(F(";\tDB\tDebug mode toggle"));
    Serial.println(F(";\t00\tRounding mode toggle"));
    Serial.println(F(";\tL:\tNew Location"));
    Serial.println(F(";\tWW\tWrite Calibration data to EEPROM"));
    Serial.println(F(";\tW+\tOverwrite Backup Calibration data"));
    Serial.println(F(";\tW-\tRestore from Backup Calibration data"));
    Serial.println(F(";\tE+\tSet Flag to send next run to EEPROM"));
    Serial.println(F(";\tE-\tClear Flag to send next run to EEPROM"));
    Serial.println(F(";\tEC\tClear EEPROM Storage"));
    Serial.println(F(";\tED\tDump data stored in EEPROM"));
    Serial.println(F(";\tLL\tList implemented commands"));
    Serial.println(F(";\t??\tList implemented commands"));
    Serial.println(F(";\tSS\tShutdown (send twice)"));
    Serial.println(F(";\t!!\tReset (send twice)"));
    PrintSeperatorLine();    
    // example of application specific command implimneted
    // these two commands write test data to the EEPROM working storage
    Serial.println(F(";\tZ1\tWrite test data 1"));
    Serial.println(F(";\tZ2\tWrite test data 2"));
    // special calibration mode
    Serial.println(F(";\tZZ\tToggle 5 Second reporting"));
    // Dump all EEPROM memory to Serial in Hex and ASCII
    Serial.println(F(";\tZD\tDump ALLL EEPROM to serial"));
    PrintSeperatorLine();    
    Serial.println(F(";\t\tResponse 'XX' = not implemented"));    
    Serial.println(F(";\t\tResponse '??' = not recognized"));
    PrintSeperatorLine();    
  }
  
//----------------------------------------------------------------- 
void PrintSeperatorLine()
  { Serial.print("; ");
    for (byte i=0; i<36; i++) Serial.print('-');
    Serial.println();
  }

//================================================================  
boolean ReadTwoCharacters()
 { char c1=0,c2=0,c3=-1;
   byte m=0;
   boolean EOC=true;  // End of Command Terminator
   boolean OurReturn=false;
      
   // It is not to be believed how much effort went into creating this simple function to read 
   // two characters. I noted a bit of problem reading characters from the serail port when 
   // the loop was too fast therefore I have added a bit of a delay to insure the serial port 
   // library can keep up. Worst case senario this function can take more than 250 milliseconds.
   // Normally when this functionis called we expect the htree bytes we need to be in the buffer 
   // but if there is noise on the line or a parrot randomly pecking at the keyboard it could 
   // take a bit longer.
   // 
   // by defintion we are looking for two characters followed by a terminator
   // we define a command terminatore to be a carriage return, new line or null character
   // --- for good measure we are including the tab character and space as well
   // space was added because it is impossible to send a tab character from the Ardunion IDE
   // we will accept any combination of those characters as a single terminator
   // we will accept the last two printable ASCII characters before a terminator for our command
   // we keep reading until we get a terminator, but we will only read for a short period
   
   // but before we do anything else we are going to save the rpevious command for posterity
   prevcmd[0]=cmd[0];            // actually we are saving it so that shutdown 
   prevcmd[1]=cmd[1];            // and reset can check it before they execute
   
   while ((c3 != 13) && (c3 != 10) && (c3 != 9) && (c3 != 0) && (c3 != 32) && (m<25))
     { // if we have a valid ASCII character for c3 then roll the charaters down
       if (Serial.available()>0) 
         { c3=Serial.read();
           if (c3>32) {c1=c2; c2=c3;}
         }  
       // we need a bit of a delay to let the serial interface catch up
       // after 25 empty reads we give up
       else { delay (10); m++;}
     }
   // DebugPrintCharacters (c1,c2,c3,m);  
   
   // we are very liberal about what we will accept for a command terminator 
   // but we insist on having one.  
   if ((c3 != 13) && (c3 != 10) && (c3 != 9) && (c3 != 0) && (c3 != 32)) EOC= false;
   // we need to drain any remaining command terminator characters from the serail buffer
   else DrainCmdTermiantors();
   
   // now check for valid ASCII charaters and End of Line
   if ((c1>32) && (c2>32) && EOC)
     { // OK... we have something to work with
       // Convert lower case to UPPER case excpet "w"
       // DebugPrintCharacters (c1,c2,c3);
       if ((c1 != 'w') && (c1 >96) && (c1 <123)) c1 = (c1 -32);
       if ((c2 != 'w') && (c2 >96) && (c2 <123)) c2 = (c2 -32);
       // DebugPrintCharacters (c1,c2);
       cmd[0]=c1;
       cmd[1]=c2;
       OurReturn=true;
     }
   // whatever it was that was sent did not meet our criteria
   // inform the parrot that he or she must do better  
   else Serial.println(F("; ?? ??"));
   return OurReturn;
 }
 
//----------------------------------------------------------------
void DrainCmdTermiantors()
 { char c3=0;
   // removed leading command terminators from serial buffer 
   delay (10); c3=Serial.peek();
   while ((c3==13) || (c3==10) || (c3==9) || (c3==0) || (c3 == 32)) 
     { c3=Serial.read(); 
       delay (10); 
       c3=Serial.peek();
     }
   // c3 should at this point should be -1 unless there are more commands/charaters in the buffer
 }

//================================================================ 
// overloaded debugging function for debugging the above input routine
void DebugPrintCharacters (char c1, char c2, char c3, byte m)
  {
    if (DeBug == true)
       { Serial.print   ("Received: ");
         Serial.print   (c1);
         // Serial.print   (" ");
         Serial.print   (c2);
         Serial.print   (" ");
         if (c3 != 0) 
           { Serial.print (c3, DEC);
             Serial.print   (" ");
           }
         if (m != 0) Serial.print (m, DEC);
         Serial.println ();
       }
  } 
void DebugPrintCharacters (char c1, char c2, char c3)
  { byte m=0;
    DebugPrintCharacters (c1,c2,c3,m);
  }  
void DebugPrintCharacters (char c1, char c2)
  { byte m=0;
    char c3=0;
    DebugPrintCharacters (c1,c2,c3,m);
  } 
 
//----------------End of Main File------------------------------  

